"""
Sprite class for encapsulating all visible objects in ggame applications.
"""
import os
import math
from ggame.sysdeps import GFX_Sprite
from ggame.asset import (
    RectangleAsset,
    CircleAsset,
    ImageAsset,
    PolygonAsset,
    EllipseAsset,
    TextAsset,
    LineAsset,
)
from ggame.app import App


# pylint: disable=useless-object-inheritance
class Sprite(object):  # pylint: disable=too-many-public-methods
    """
    The Sprite class combines the idea of a visual/graphical asset, a
    position on the screen, and *behavior*. Although the Sprite can be
    used as-is, it is generally subclassed to give it some desired behavior.

    When subclassing the Sprite class, you may customize the initialization
    code to use a specific asset. A 'step' or 'poll' method may be added
    for handling per-frame actions (e.g. checking for collisions). Step or poll
    functions are not automatically called by the :class:`~ggame.app.App` class,
    but you may subclass the :class:`~ggame.app.App` class in order to do this.

    Furthermore, you may wish to define event callback methods in your customized
    sprite class. With customized creation, event handling, and periodic processing
    you can achieve fully autonomous behavior for your sprite objects.

    :param asset asset: An existing graphical asset

    :param tuple(int,int) pos:  The sprite position may be provided, which
        specifies the starting (x,y) coordinates of the sprite on the screen.
        By default, the position of a sprite defines the location of its upper-left
        hand corner. This behavior can be modified by customizing its
        :data:`center`.

    :param asset edgedef: An edge definition asset may be provided, which
        specifies an asset that will be used to define the boundaries of
        the sprite for the purpose of collision detection. If no `edgedef`
        asset is given, the required asset is used, which will be a rectangular
        asset in the case of an image texture. This option is typically used
        to define a visible image outline for a texture-based sprite that has
        a transparent texture image background.

    :returns: Nothing. If the position is on screen the sprite will be displayed
        in the browser.

    Example of use:

    .. literalinclude:: ../examples/spritesprite.py

    This creates a sprite using the 'player.png' image, positioned with its
    upper-left corner at coordinates (100,100) and with a 50 pixel radius
    circular collision border.

    """

    def __init__(self, asset, pos=(0, 0), edgedef=None):
        self._index = 0
        if isinstance(asset, ImageAsset):
            self.asset = asset
            try:
                self.gfx = GFX_Sprite(asset.gfx)  # gfx is PIXI Sprite
            except:  # pylint: disable=bare-except
                self.gfx = None
                raise
        elif isinstance(
            asset, (RectangleAsset, CircleAsset, EllipseAsset, PolygonAsset, LineAsset)
        ):
            self.asset = asset
            self.gfx = GFX_Sprite(asset.gfx.generateTexture())
        elif isinstance(asset, TextAsset):
            self.asset = asset.clone()
            self.gfx = self.asset.gfx  # gfx is PIXI Text (from Sprite)
            self.gfx.visible = True
        if not edgedef:
            self.edgedef = asset
        else:
            self.edgedef = edgedef
        self.xmin = self.xmax = self.ymin = self.ymax = 0
        self.position = pos
        """Tuple indicates the position of the sprite on the screen."""
        self._extentsdirty = True
        """Boolean indicates if extents must be calculated before collision test"""
        self._createBaseVertices()
        self._absolutevertices = None
        self.setExtents()
        App.add(self)

    def _createBaseVertices(self):
        """
        Create sprite-relative list of vertex coordinates for boundary
        """
        self._basevertices = []
        assettype = type(self.edgedef)
        if assettype in [RectangleAsset, ImageAsset, TextAsset]:
            self._basevertices = [
                (0, 0),
                (0, self.edgedef.height),
                (self.edgedef.width, self.edgedef.height),
                (self.edgedef.width, 0),
            ]
        elif assettype in [PolygonAsset, LineAsset]:
            if assettype is PolygonAsset:
                self._basevertices = self.edgedef.path[:-1]
            elif assettype is LineAsset:
                self._basevertices = [
                    (0, 0),
                    (self.edgedef.delta_x, self.edgedef.delta_y),
                ]
            xpoints, ypoints = zip(*self._basevertices)
            xmin = min(xpoints)
            ymin = min(ypoints)
            xpoints = [x - xmin for x in xpoints]
            ypoints = [y - ymin for y in ypoints]
            self._basevertices = list(zip(xpoints, ypoints))
        elif assettype is EllipseAsset:
            w = self.edgedef.halfw * 2
            h = self.edgedef.halfh * 2
            self._basevertices = [(0, 0), (0, h), (w, h), (w, 0)]

    def _xformVertices(self):
        """
        Create window-relative list of vertex coordinates for boundary
        """
        # find center as sprite-relative points (note sprite may be scaled)
        x = self.width * self.fxcenter / self.scale
        y = self.height * self.fycenter / self.scale
        if self.scale != 1.0:
            sc = self.scale
            # center-relative, scaled coordinates
            crsc = [((xp - x) * sc, (yp - y) * sc) for xp, yp in self._basevertices]
        else:
            crsc = [(xp - x, yp - y) for xp, yp in self._basevertices]

        # absolute, rotated coordinates
        c = math.cos(self.rotation)
        s = math.sin(self.rotation)
        self._absolutevertices = [
            (self.x + x * c + y * s, self.y + -x * s + y * c) for x, y in crsc
        ]

    def setExtents(self):
        """
        update min/max x and y based on position, center, width, height
        """
        if self._extentsdirty:
            if isinstance(self.edgedef, CircleAsset):
                th = (
                    math.atan2(self.fycenter - 0.5, 0.5 - self.fxcenter) + self.rotation
                )
                d = self.edgedef.radius * 2 * self.scale
                l = (
                    math.sqrt(
                        math.pow(self.fxcenter - 0.5, 2)
                        + math.pow(self.fycenter - 0.5, 2)
                    )
                    * d
                )
                self.xmin = self.x + int(l * math.cos(th)) - d // 2
                self.ymin = self.y - int(l * math.sin(th)) - d // 2
                self.xmax = self.xmin + d
                self.ymax = self.ymin + d
            else:
                # Build vertex list
                self._xformVertices()
                x, y = zip(*self._absolutevertices)
                self.xmin = min(x)
                self.xmax = max(x)
                self.ymin = min(y)
                self.ymax = max(y)
            self._extentsdirty = False

    def firstImage(self):
        """
        Select and display the *first* image used by this sprite. This only
        does something useful if the asset is an :class:`~ggame.asset.ImageAsset`
        defined with multiple images.
        """
        self.gfx.texture = self.asset[0]

    def lastImage(self):
        """
        Select and display the *last* image used by this sprite. This only
        does something useful if the asset is an :class:`~ggame.asset.ImageAsset`
        defined with multiple images.
        """
        self.gfx.texture = self.asset[-1]

    def nextImage(self, wrap=False):
        """
        Select and display the *next* image used by this sprite.
        If the current image is already the *last* image, then
        the image is not advanced.

        :param boolean wrap: If `True`, then calling
            :meth:`nextImage` on the last image will cause the *first*
            image to be loaded.

        This only does something useful if the asset is an
        :class:`~ggame.asset.ImageAsset` defined with multiple images.
        """
        self._index += 1
        if self._index >= len(self.asset):
            if wrap:
                self._index = 0
            else:
                self._index = len(self.asset) - 1
        self.gfx.texture = self.asset[self._index]

    def prevImage(self, wrap=False):
        """
        Select and display the *previous* image used by this sprite.
        If the current image is already the *first* image, then
        the image is not changed.

        :param boolean wrap: If `True`, then calling
            :meth:`prevImage` on the first image will cause the *last*
            image to be loaded.

        This only does something useful if the asset is an
        :class:`~ggame.asset.ImageAsset` defined with multiple images.
        """
        self._index -= 1
        if self._index < 0:
            if wrap:
                self._index = len(self.asset) - 1
            else:
                self._index = 0
        self.gfx.texture = self.asset[self._index]

    def setImage(self, index=0):
        """
        Select the image to display by giving its `index`.

        :param int index: An index to specify the image to display.
            A value of zero represents the *first* image in the asset.

        This is equivalent to setting the :data:`index` property
        directly.

        This only does something useful if the asset is an
        :class:`~ggame.asset.ImageAsset` defined with multiple images.
        """
        self.index = index

    def rectangularCollisionModel(self):
        """
        Obsolete. No op.
        """

    def circularCollisionModel(self):
        """
        Obsolete. No op.
        """

    @property
    def index(self):
        """
        This is an integer index into the list of images available for this sprite.
        """
        return self._index

    @index.setter
    def index(self, value):
        self._index = value
        try:
            self.gfx.texture = self.asset[self._index]
        except:  # pylint: disable=bare-except
            self._index = 0
            self.gfx.texture = self.asset[self._index]

    @property
    def width(self):
        """
        This is an integer representing the display width of the sprite.
        Assigning a value to the width will scale the image horizontally.
        """
        return self.gfx.width

    @width.setter
    def width(self, value):
        self.gfx.width = value
        self._extentsdirty = True

    @property
    def height(self):
        """
        This is an integer representing the display height of the sprite.
        Assigning a value to the height will scale the image vertically.
        """
        return self.gfx.height

    @height.setter
    def height(self, value):
        self.gfx.height = value
        self._extentsdirty = True

    @property
    def x(self):
        """
        This represents the x-coordinate of the sprite on the screen. Assigning
        a value to this attribute will move the sprite horizontally.
        """
        return self.gfx.position.x

    @x.setter
    def x(self, value):
        delta_x = value - self.gfx.position.x
        self.xmax += delta_x
        self.xmin += delta_x
        # Adjust extents directly with low overhead
        self.gfx.position.x = value

    @property
    def y(self):
        """
        This represents the y-coordinate of the sprite on the screen. Assigning
        a value to this attribute will move the sprite vertically.
        """
        return self.gfx.position.y

    @y.setter
    def y(self, value):
        delta_y = value - self.gfx.position.y
        self.ymax += delta_y
        self.ymin += delta_y
        # Adjust extents directly with low overhead
        self.gfx.position.y = value

    @property
    def position(self):
        """
        This represents the (x,y) coordinates of the sprite on the screen. Assigning
        a value to this attribute will move the sprite to the new coordinates.
        """
        return (self.gfx.position.x, self.gfx.position.y)

    @position.setter
    def position(self, value):
        self.x, self.y = value

    @property
    def fxcenter(self):
        """
        This represents the horizontal position of the sprite "center", as a floating
        point number between 0.0 and 1.0. A value of 0.0 means that the x-coordinate
        of the sprite refers to its left hand edge. A value of 1.0 refers to its
        right hand edge. Any value in between may be specified. Values may be assigned
        to this attribute.
        """
        try:
            return self.gfx.anchor.x
        except:  # pylint: disable=bare-except
            return 0.0

    @fxcenter.setter
    def fxcenter(self, value):
        try:
            self.gfx.anchor.x = value
            self._extentsdirty = True
        except:  # pylint: disable=bare-except
            pass

    @property
    def fycenter(self):
        """
        This represents the vertical position of the sprite "center", as a floating
        point number between 0.0 and 1.0. A value of 0.0 means that the x-coordinate
        of the sprite refers to its top edge. A value of 1.0 refers to its
        bottom edge. Any value in between may be specified. Values may be assigned
        to this attribute.
        """
        try:
            return self.gfx.anchor.y
        except:  # pylint: disable=bare-except
            return 0.0

    @fycenter.setter
    def fycenter(self, value):
        try:
            self.gfx.anchor.y = value
            self._extentsdirty = True
        except:  # pylint: disable=bare-except
            pass

    @property
    def center(self):
        """
        This attribute represents the horizontal and vertical position of the
        sprite "center" as a tuple of floating point numbers. See the
        descriptions for :data:`fxcenter` and :data:`fycenter` for
        more details.
        """
        try:
            return (self.gfx.anchor.x, self.gfx.anchor.y)
        except:  # pylint: disable=bare-except
            return (0.0, 0.0)

    @center.setter
    def center(self, value):
        try:
            self.gfx.anchor.x = value[0]
            self.gfx.anchor.y = value[1]
            self._extentsdirty = True
        except:  # pylint: disable=bare-except
            pass

    @property
    def visible(self):
        """
        This boolean attribute may be used to change the visibility of the sprite.
        Setting `~ggame.Sprite.visible` to `False` will prevent the sprite from
        rendering on the screen.
        """
        return self.gfx.visible

    @visible.setter
    def visible(self, value):
        self.gfx.visible = value

    @property
    def scale(self):
        """
        This attribute may be used to change the size of the sprite ('scale' it) on the
        screen. Value may be a floating point number. A value of 1.0 means that the
        sprite image will keep its original size. A value of 2.0 would double it, etc.
        """
        try:
            return self.gfx.scale.x
        except AttributeError:
            return 1.0

    @scale.setter
    def scale(self, value):
        self.gfx.scale.x = value
        self.gfx.scale.y = value
        self._extentsdirty = True

    @property
    def rotation(self):
        """
        This attribute may be used to change the rotation of the sprite on the screen.
        Value may be a floating point number. A value of 0.0 means no rotation. A value
        of 1.0 means  a rotation of 1 radian in a counter-clockwise direction. One
        radian is 180/pi or approximately 57.3 degrees.
        """
        try:
            return -self.gfx.rotation
        except AttributeError:
            return 0.0

    @rotation.setter
    def rotation(self, value):
        if self.gfx.rotation != -value:
            self.gfx.rotation = -value
            self._extentsdirty = True

    @classmethod
    def collidingCircleWithPoly(cls, circ, poly):  # pylint: disable=unused-argument
        """
        Determine if a CircleAsset sprite overlaps with a PolygonAsset sprite. This
        method is called after determining that the two objects are overlapping in their
        overall extents.

        :param Sprite circ: A CircleAsset-based sprite.
        :param Sprite poly: A PolygonAsset-based sprite.
        :returns: True if the sprites are overlapping, False otherwise.
        :rtype: boolean
        """
        return True  # no implementation yet

    def collidingPolyWithPoly(self, obj):
        """
        Determine if a pair of PolygonAsset-based sprites are overlapping. This
        method is called after determining that the two objects are overlapping in their
        overall extents. This should onlyb e called if `self` is a PolygonAsset-based
        sprite.

        :param Sprite obj: A PolygonAsset-based sprite.
        :returns: True if slef overlaps with obj, False otherwise.
        :rtype: boolean
        """
        return True  # no implementation yet

    def collidingWith(self, obj):
        """
        Determine if this sprite is currently overlapping another
        sprite object.

        :param Sprite obj: A reference to another Sprite object.

        :rtype: boolean

        :returns: `True` if this the sprites are overlapping, `False` otherwise.
        """
        if self is obj:
            return False
        self.setExtents()
        obj.setExtents()
        # Gross check for overlap will usually rule out a collision
        if (
            self.xmin > obj.xmax
            or self.xmax < obj.xmin
            or self.ymin > obj.ymax
            or self.ymax < obj.ymin
        ):
            return False
        # Otherwise, perform a careful overlap determination
        if isinstance(self.asset, CircleAsset):
            if isinstance(obj.asset, CircleAsset):
                # two circles .. check distance between
                sx = (self.xmin + self.xmax) / 2
                sy = (self.ymin + self.ymax) / 2
                ox = (obj.xmin + obj.xmax) / 2
                oy = (obj.ymin + obj.ymax) / 2
                d = math.sqrt((sx - ox) ** 2 + (sy - oy) ** 2)
                return d <= self.width / 2 + obj.width / 2
            return self.collidingCircleWithPoly(self, obj)
        if isinstance(obj.asset, CircleAsset):
            return self.collidingCircleWithPoly(obj, self)
        return self.collidingPolyWithPoly(obj)

    def collidingWithSprites(self, sclass=None):
        """
        Determine if this sprite is colliding with any other sprites
        of a certain class.

        :param class sclass: A class identifier that is either :class:`Sprite`
            or a subclass of it that identifies the class of sprites to check
            for collisions. If `None` then all objects that are subclassed from
            the :class:`Sprite` class are checked.

        :rtype: list

        :returns: A (potentially empty) list of sprite objects of the given
            class that are overlapping with this sprite.
        """
        if sclass is None:
            slist = App.spritelist
        else:
            slist = App.getSpritesbyClass(sclass)
        return list(filter(self.collidingWith, slist))

    @staticmethod
    def getImagePath(imagename):
        """
        Determine a path to the ggame-provided image that will work for both online
        (runpython.org) and locally installed ggame libraries. Do not use this with
        user-provided images.

        :param str imagename: The name of an image file found inside the
            ggame/images folder.
        """
        # differences in online vs. local operation
        try:
            thispath = os.path.dirname(__file__)
            imagepath = os.path.join(thispath, "images")
        except NameError:
            imagepath = "images"
        return os.path.join(imagepath, imagename)

    def destroy(self):
        """
        Prevent the sprite from being displayed or checked in collision
        detection. Once this is called, the sprite can no longer be displayed
        or used. If you only want to prevent a sprite from being displayed,
        set the :data:`visible` attribute to `False`.
        """
        try:
            App.remove(self)
            self.gfx.destroy()
        except ValueError:
            pass



# pylint: skip-file


def module_exists(module_name):
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        return True


if module_exists("browser"):
    from browser import window, document, load

    major = window.__BRYTHON__.implementation[0]
    minor = window.__BRYTHON__.implementation[1]
    GFX = window.PIXI
    GFX_Rectangle = GFX.Rectangle.new
    GFX_Texture = GFX.Texture.new
    GFX_Texture_fromImage = GFX.Texture.fromImage.new
    GFX_Sprite = GFX.Sprite.new
    GFX_Graphics = GFX.Graphics.new()
    GFX_Text = GFX.Text.new
    GFX_NewStage = GFX.Container.new
    SND = window.buzz
    SND_Sound = SND.sound.new
    GFX_DetectRenderer = GFX.autoDetectRenderer

    class GFX_Window(object):
        def __init__(self, width, height, onclose):
            canvas = window.document.getElementById("ggame-canvas")
            if canvas:
                self._w = window
                window.bsUI.graphicsmode()
                options = {"transparent": True, "antialias": True, "view": canvas}
                attachpoint = window.document.getElementById("graphics-column")
                w, h = attachpoint.clientWidth, attachpoint.clientHeight
            else:
                self._w = window.open("", "")
                w, h = self._w.innerWidth * 0.9, self._w.innerHeight * 0.9
                options = {"transparent": True, "antialias": True}
                attachpoint = self._w.document.body
            GFX.utils._saidHello = True
            # ugly hack to block pixi banner
            self._stage = GFX_NewStage()
            self.width = width if width != 0 else int(w)
            self.height = height if height != 0 else int(h)
            self.renderer = GFX.autoDetectRenderer(self.width, self.height, options)
            attachpoint.appendChild(self.renderer.view)
            self._w.ggame_quit = onclose

        def bind(self, evtspec, callback):
            self._w.document.body.unbind(evtspec)
            self._w.document.body.bind(evtspec, callback)

        def unbind(self, evtspec):
            self._w.document.body.unbind(evtspec)

        def add(self, obj):
            self._stage.addChild(obj)

        def remove(self, obj):
            self._stage.removeChild(obj)

        def animate(self, stepcallback):
            self.renderer.render(self._stage)
            self._w.requestAnimationFrame(stepcallback)

        def destroy(self):
            SND.all().stop()
            self.renderer.destroy()

elif module_exists("pygame"):
    try:
        from ggame.pygamedeps import *
    except ImportError:
        from pygamedeps import *

else:
    try:
        from ggame.headlessdeps import *
    except ImportError:
        from headlessdeps import *



"""
These mathematics and geometry extensions subclass the :class:`~ggame.app.App`
and :class:`~ggame.sprite.Sprite` classes to create a framework for building
apps that mimic some of the functionality of online math tools like Geogebra.

This :mod:`~ggame.mathapp` module implements base classes for
:class:`~ggame.sprite.Sprite`-based classes defined in this module.

These extensions are very experimental and are not fully developed!
"""

from abc import ABCMeta, abstractmethod
from time import time
from math import sqrt
from collections import namedtuple
from ggame.sprite import Sprite
from ggame.asset import Color, LineStyle, ImageAsset
from ggame.app import App


class MathApp(App):
    """
    MathApp is a subclass of the ggame :class:`~ggame.app.App` class. It
    incorporates the following extensions:

    * Support for zooming the display using the mouse wheel
    * Support for click-dragging the display using the mouse button
    * Automatic execution of step functions in all objects and sprites
        sub-classed from :class:`_MathDynamic`.

    :param float scale: Optional parameter sets the initial scale of the
        display in units of pixels per logical unit. The default is 200.

    :returns: MathApp instance
    """

    _DEFAULTSCALE = 200
    scale = _DEFAULTSCALE  # pixels per unit
    _xcenter = 0  # center of screen in units
    _ycenter = 0
    _mathVisualList = []
    _mathDynamicList = []
    _mathMovableList = []
    _mathSelectableList = []
    _mathStrokableList = []
    _viewNotificationList = []
    time = 0

    def __init__(self, scale=_DEFAULTSCALE):
        MathApp.time = 0
        self._starttime = time()
        super().__init__()
        MathApp.scale = scale  # pixels per unit
        # register event callbacks
        self.listenMouseEvent("click", self._handleMouseClick)
        self.listenMouseEvent("mousedown", self._handleMouseDown)
        self.listenMouseEvent("mouseup", self._handleMouseUp)
        self.listenMouseEvent("mousemove", self._handleMouseMove)
        self.listenMouseEvent("wheel", self._handleMouseWheel)
        self._mousedown = False
        self._mouse_captured_object = None
        self._mouse_stroked_object = None
        self._mouse_down_object = None
        self._mouse_x = self._mouse_y = None
        self._touchAllVisuals()
        self._selectedobj = None

    def step(self):
        """
        The step method overrides :func:`~ggame.app.App.step` in the
        :class:`~ggame.app.App` class, executing step functions in all
        objects subclassed from :class:`_MathDynamic`.
        """
        MathApp.time = time() - self._starttime
        for spr in self._mathDynamicList:
            spr.step()

    def _touchAllVisuals(self):
        # touch all visual object assets to use scaling
        for obj in self._mathVisualList:
            obj.touchAsset(True)

    @classmethod
    def logicalToPhysical(cls, lp):
        """
        Transform screen coordinates from logical to physical space. Output
        depends on the current 'zoom' and 'pan' of the screen.

        :param tuple(float,float) lp: Logical screen coordinates (x, y)

        :rtype: tuple(float,float)

        :returns: Physical screen coordinates (x, y)
        """

        def xxform(xvalue, xscale, xcenter, physwidth):
            """X transform"""
            return int((xvalue - xcenter) * xscale + physwidth / 2)

        def yxform(yvalue, yscale, ycenter, physheight):
            """Y transform"""
            return int(physheight / 2 - (yvalue - ycenter) * yscale)

        try:
            return (
                xxform(lp[0], cls.scale, cls._xcenter, cls.win.width),
                yxform(lp[1], cls.scale, cls._ycenter, cls.win.height),
            )
        except AttributeError:
            return lp

    @classmethod
    def physicalToLogical(cls, pp):
        """
        Transform screen coordinates from physical to logical space. Output
        depends on the current 'zoom' and 'pan' of the screen.

        :param tuple(float,float) lp: Physical screen coordinates (x, y)

        :rtype: tuple(float,float)

        :returns: Logical screen coordinates (x, y)
        """

        def xxform(xvalue, xscale, xcenter, physwidth):
            """X transform"""
            return (xvalue - physwidth / 2) / xscale + xcenter

        def yxform(yvalue, yscale, ycenter, physheight):
            """Y transform"""
            return (physheight / 2 - yvalue) / yscale + ycenter

        try:
            return (
                xxform(pp[0], cls.scale, cls._xcenter, cls.win.width),
                yxform(pp[1], cls.scale, cls._ycenter, cls.win.height),
            )
        except AttributeError:
            return pp

    @classmethod
    def translateLogicalToPhysical(cls, pp):
        """
        Transform screen translation from logical to physical space. Output
        only depends on the current 'zoom' of the screen.

        :param tuple(float,float) lp: Logical screen translation pair
            (delta x, delta y)

        :rtype: tuple(float,float)

        :returns: Physical screen translation ordered pair (delta x, delta y)
        """

        def xxform(xvalue, xscale):
            """X transform"""
            return xvalue * xscale

        def yxform(yvalue, yscale):
            """Y transform"""
            return -yvalue * yscale

        try:
            return (xxform(pp[0], cls.scale), yxform(pp[1], cls.scale))
        except AttributeError:
            return pp

    @classmethod
    def translatePhysicalToLogical(cls, pp):
        """
        Transform screen translation from physical to logical space. Output
        only depends on the current 'zoom' of the screen.

        :param tuple(float,float) lp: Physical screen translation pair
            (delta x, delta y)

        :rtype: tuple(float,float)

        :returns: Logical screen translation ordered pair (delta x, delta y)
        """

        def xxform(xvalue, xscale):
            """X transform"""
            return xvalue / xscale

        def yxform(yvalue, yscale):
            """Y transform"""
            return -yvalue / yscale

        try:
            return (xxform(pp[0], cls.scale), yxform(pp[1], cls.scale))
        except AttributeError:
            return pp

    def _handleMouseClick(self, event):
        found = False
        for obj in self._mathSelectableList:
            if obj.physicalPointTouching((event.x, event.y)):
                found = True
                if not obj.selected:
                    obj.select()
                    self._selectedobj = obj
        if not found and self._selectedobj:
            self._selectedobj.unselect()
            self._selectedobj = None

    def _handleMouseDown(self, event):
        self._mousedown = True
        self._mouse_captured_object = None
        self._mouse_stroked_object = None
        for obj in self._mathSelectableList:
            if obj.physicalPointTouching((event.x, event.y)):
                obj.mousedown()
                self._mouse_down_object = obj
                break
        for obj in self._mathMovableList:
            if obj.physicalPointTouching((event.x, event.y)) and not (
                obj.strokable and obj.canstroke((event.x, event.y))
            ):
                self._mouse_captured_object = obj
                break
        if not self._mouse_captured_object:
            for obj in self._mathStrokableList:
                if obj.canstroke((event.x, event.y)):
                    self._mouse_stroked_object = obj
                    break

    def _handleMouseUp(self, _event):
        if self._mouse_down_object:
            self._mouse_down_object.mouseup()
            self._mouse_down_object = None
        self._mousedown = False
        self._mouse_captured_object = None
        self._mouse_stroked_object = None

    def _handleMouseMove(self, event):
        if not self._mouse_x:
            self._mouse_x = event.x
            self._mouse_y = event.y
        dx = event.x - self._mouse_x
        dy = event.y - self._mouse_y
        self._mouse_x = event.x
        self._mouse_y = event.y
        if self._mousedown:
            if self._mouse_captured_object:
                self._mouse_captured_object.translate((dx, dy))
            elif self._mouse_stroked_object:
                self._mouse_stroked_object.stroke(
                    (self._mouse_x, self._mouse_y), (dx, dy)
                )
            else:
                lmove = self.translatePhysicalToLogical((dx, dy))
                MathApp._xcenter -= lmove[0]
                MathApp._ycenter -= lmove[1]
                self._touchAllVisuals()
                self._viewNotify("translate")

    def _handleMouseWheel(self, event):
        zoomfactor = event.wheeldelta / 100
        zoomfactor = 1 + zoomfactor if zoomfactor > 0 else 1 + zoomfactor
        if zoomfactor > 1.2:
            zoomfactor = 1.2
        elif zoomfactor < 0.8:
            zoomfactor = 0.8
        MathApp.scale *= zoomfactor
        self._touchAllVisuals()
        self._viewNotify("zoom")

    @property
    def view_position(self):
        """
        Attribute is used to get or set the current logical coordinates
        at the center of the screen as a tuple of floats (x,y).
        """
        return (MathApp._xcenter, MathApp._ycenter)

    @view_position.setter
    def view_position(self, pos):
        MathApp._xcenter, MathApp._ycenter = pos
        self._touchAllVisuals()
        self._viewNotify("translate")

    @classmethod
    def addViewNotification(cls, handler):
        """
        Register a function or method to be called in the event the view
        position or zoom changes.

        :param function handler: The function or method to be called
        :returns: Nothing
        """
        cls._viewNotificationList.append(handler)

    @classmethod
    def removeViewNotification(cls, handler):
        """
        Remove a function or method from the list of functions to be called
        in the event of a view position or zoom change.

        :param function handler: The function or method to be removed
        :returns: Nothing
        """
        cls._viewNotificationList.remove(handler)

    def _viewNotify(self, viewchange):
        for handler in self._viewNotificationList:
            handler(
                viewchange=viewchange,
                scale=self.scale,
                center=(self._xcenter, self._ycenter),
            )

    @classmethod
    def distance(cls, pos1, pos2):
        """
        Utility for calculating the distance between any two points.

        :param tuple(float,float) pos1: The first point
        :param tuple(float,float) pos2: The second point
        :rtype: float
        :returns: The distance between the two points (using Pythagoras)
        """
        return sqrt((pos2[0] - pos1[0]) ** 2 + (pos2[1] - pos1[1]) ** 2)

    @classmethod
    def addVisual(cls, obj):
        """
        Add a visual object to the visual object list.

        :param object obj: The object to add
        :returns: None
        """
        if isinstance(obj, _MathVisual):
            cls._mathVisualList.append(obj)

    @classmethod
    def removeVisual(cls, obj):
        """
        Remove a visual object from the visual object list.

        :param object obj: The object to remove
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj in cls._mathVisualList:
            cls._mathVisualList.remove(obj)

    @classmethod
    def addDynamic(cls, obj):
        """
        Add a dynamic object to the dynamic object list.

        :param object obj: The object to add
        :returns: None
        """
        if isinstance(obj, _MathDynamic) and obj not in cls._mathDynamicList:
            cls._mathDynamicList.append(obj)

    @classmethod
    def removeDynamic(cls, obj):
        """
        Remove a dynamic object from the dynamic object list.

        :param object obj: The object to remove
        :returns: None
        """
        if isinstance(obj, _MathDynamic) and obj in cls._mathDynamicList:
            cls._mathDynamicList.remove(obj)

    @classmethod
    def addMovable(cls, obj):
        """
        Add a movable object to the movable object list.

        :param object obj: The object to add
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj not in cls._mathMovableList:
            cls._mathMovableList.append(obj)

    @classmethod
    def removeMovable(cls, obj):
        """
        Remove a movable object from the movable object list.

        :param object obj: The object to remove
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj in cls._mathMovableList:
            cls._mathMovableList.remove(obj)

    @classmethod
    def addSelectable(cls, obj):
        """
        Add a selectable object to the selectable object list.

        :param object obj: The object to add
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj not in cls._mathSelectableList:
            cls._mathSelectableList.append(obj)

    @classmethod
    def removeSelectable(cls, obj):
        """
        Remove a selectable object from the selectable object list.

        :param object obj: The object to remove
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj in cls._mathSelectableList:
            cls._mathSelectableList.remove(obj)

    @classmethod
    def addStrokable(cls, obj):
        """
        Add a strokable object to the strokable object list.

        :param object obj: The object to add
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj not in cls._mathStrokableList:
            cls._mathStrokableList.append(obj)

    @classmethod
    def removeStrokable(cls, obj):
        """
        Remove a strokable object from the strokable object list.

        :param object obj: The object to remove
        :returns: None
        """
        if isinstance(obj, _MathVisual) and obj in cls._mathStrokableList:
            cls._mathStrokableList.remove(obj)

    @classmethod
    def destroy(cls):
        """
        This will clean up any class level storage.
        """
        App.destroy()  # hit the App class first
        MathApp.time = 0
        MathApp._mathVisualList = []
        MathApp._mathDynamicList = []
        MathApp._mathMovableList = []
        MathApp._mathSelectableList = []
        MathApp._mathStrokableList = []
        MathApp._viewNotificationList = []


class _MathDynamic(metaclass=ABCMeta):
    def __init__(self):
        self._dynamic = False  # not switched on, by default!

    def destroy(self):
        """
        Destroy resources, if any and remove from global lists.
        """
        MathApp.removeDynamic(self)

    def step(self):
        """
        Override in your child class to perform periodic processing.
        """

    def eval(self, val):
        """
        Evaluate a potentially callable argument, returning a callable
        object.

        :param val: A simple variable or function reference that can be
            called without any argument, returning a value.
        :rtype: function
        :returns: A function that can be called to retrieve the value passed
            in.
        """
        if callable(val):
            self._setDynamic()  # dynamically defined .. must step
            return val
        return lambda: val

    def _setDynamic(self):
        MathApp.addDynamic(self)
        self._dynamic = True


class _MathVisual(Sprite, _MathDynamic, metaclass=ABCMeta):
    """
    Abstract Base Class for all visual, potentially dynamic objects.

    :param Asset asset: A valid ggame asset object.

    :param list args: A list of required positional or non-positional arguments
        as named in the _posinputsdef and _nonposinputsdef lists overridden
        by child classes.

    :param \\**kwargs:
        See below

    :Optional Keyword Arguments:
        * **positioning** (*string*) One of 'logical' or 'physical'
        * **size** (*int*) Size of the object (in pixels)
        * **width** (*int*) Width of the object (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object
        * **style** (*LineStyle*) Valid :class:`~ggame.asset.LineStyle` object
    """

    # a list of names (string) of required positional inputs
    _posinputsdef = []
    # a list of names (string) of required non positional inputs
    _nonposinputsdef = []
    _defaultsize = 15
    _defaultwidth = 200
    _defaultcolor = Color(0, 1)
    _defaultstyle = LineStyle(1, Color(0, 1))

    def __init__(self, asset, *args, **kwargs):
        MathApp.addVisual(self)
        # Sprite.__init__(self, asset, args[0])
        _MathDynamic.__init__(self)
        self._movable = False
        self._selectable = False
        self._strokable = False
        self.selected = False
        """
        True if object is currently selected by the UI.
        """
        self.mouseisdown = False
        """
        True if object is tracking UI mouse button as down.
        """
        self._positioning = kwargs.get("positioning", "logical")
        # positional inputs
        self._pi = namedtuple("PI", self._posinputsdef)
        # nonpositional inputs
        self._npi = namedtuple("NPI", self._nonposinputsdef)
        # standard inputs (not positional)
        standardargs = ["size", "width", "color", "style"]
        self._si = namedtuple("SI", standardargs)
        # correct number of args?
        if len(args) != len(self._posinputsdef) + len(self._nonposinputsdef):
            raise TypeError("Incorrect number of parameters provided")
        self._args = args
        # generated named tuple of functions from positional inputs
        self._posinputs = self._pi(
            *[self.eval(p) for p in args][: len(self._posinputsdef)]
        )
        self._getPhysicalInputs()
        # first positional argument must be a sprite position!
        Sprite.__init__(self, asset, self._pposinputs[0])
        # generated named tuple of functions from nonpositional inputs
        if self._nonposinputsdef:
            self._nposinputs = self._npi(
                *[self.eval(p) for p in args][(-1 * len(self._nonposinputsdef)) :]
            )
        else:
            self._nposinputs = []
        self._stdinputs = self._si(
            self.eval(kwargs.get("size", self._defaultsize)),
            self.eval(kwargs.get("width", self._defaultwidth)),
            self.eval(kwargs.get("color", self._defaultcolor)),
            self.eval(kwargs.get("style", self._defaultstyle)),
        )
        self._sposinputs = self._pi(*[0] * len(self._posinputs))
        self._spposinputs = self._pi(*self._pposinputs)
        self._snposinputs = self._npi(*[0] * len(self._nposinputs))
        self._sstdinputs = self._si(*[0] * len(self._stdinputs))

    def step(self):
        self.touchAsset()

    def _saveInputs(self, inputs):
        (
            self._sposinputs,
            self._spposinputs,
            self._snposinputs,
            self._sstdinputs,
        ) = inputs

    def _getInputs(self):
        self._getPhysicalInputs()
        return (
            self._pi(*[p() for p in self._posinputs]),
            self._pi(*self._pposinputs),
            self._npi(*[p() for p in self._nposinputs]),
            self._si(*[p() for p in self._stdinputs]),
        )

    def _getPhysicalInputs(self):
        """
        Translate all positional inputs to physical
        """
        pplist = []
        if self._positioning == "logical":
            for p in self._posinputs:
                pval = p()
                try:
                    pp = MathApp.logicalToPhysical(pval)
                except AttributeError:
                    pp = MathApp.scale * pval
                pplist.append(pp)
        else:
            # already physical
            pplist = [p() for p in self._posinputs]
        self._pposinputs = self._pi(*pplist)

    def _inputsChanged(self, saved):
        return (
            self._spposinputs != saved[1]
            or self._snposinputs != saved[2]
            or self._sstdinputs != saved[3]
        )

    def destroy(self):
        MathApp.removeVisual(self)
        MathApp.removeMovable(self)
        MathApp.removeStrokable(self)
        _MathDynamic.destroy(self)
        Sprite.destroy(self)

    def _updateAsset(self, asset):
        if not isinstance(asset, ImageAsset):
            visible = self.gfx.visible
            if MathApp.win is not None:
                MathApp.win.remove(self.gfx)
                self.gfx.destroy()
            self.asset = asset
            self.gfx = self.asset.gfx
            self.gfx.visible = visible
            if MathApp.win is not None:
                MathApp.win.add(self.gfx)
        if hasattr(self._pposinputs, "pos"):
            self.position = getattr(self._pposinputs, "pos")

    @property
    def positioning(self):
        """
        Whether object was created with 'logical' or 'physical' positioning.
        """
        return self._positioning

    @positioning.setter
    def positioning(self, val):
        pass

    @property
    def movable(self):
        """
        Whether object can be moved. Set-able and get-able.
        """
        return self._movable

    @movable.setter
    def movable(self, val):
        if not self._dynamic:
            self._movable = val
            if val:
                MathApp.addMovable(self)
            else:
                MathApp.removeMovable(self)

    @property
    def selectable(self):
        """
        Whether object can be selected by the UI. Set-able and get-able.
        """
        return self._selectable

    @selectable.setter
    def selectable(self, val):
        self._selectable = val
        if val:
            MathApp.addSelectable(self)
        else:
            MathApp.removeSelectable(self)

    @property
    def strokable(self):
        """
        Whether the object supports a click-drag input from the UI mouse.
        Set-able and get-able.
        """
        return self._strokable

    @strokable.setter
    def strokable(self, val):
        self._strokable = val
        if val:
            MathApp.addStrokable(self)
        else:
            MathApp.removeStrokable(self)

    def select(self):
        """
        Place the object in a 'selected' state.

        :param: None
        :returns: None
        """
        self.selected = True

    def unselect(self):
        """
        Place the object in an 'unselected' state.

        :param: None
        :returns: None
        """
        self.selected = False

    def mousedown(self):
        """
        Inform the object of a 'mouse down' event.

        :param: None
        :returns: None
        """
        self.mouseisdown = True

    def mouseup(self):
        """
        Inform the object of a 'mouse up' event.

        :param: None
        :returns: None
        """
        self.mouseisdown = False

    def processEvent(self, event):
        """
        Inform the object of a generic ggame event.

        :param event: The ggame event object to receive and process.
        :returns: None

        This method is intended to be overridden.
        """

    @abstractmethod
    def physicalPointTouching(self, ppos):
        """
        Determine if a physical point is considered to be touching this object.

        :param tuple(int,int) ppos: Physical screen coordinates.
        :rtype: boolean
        :returns: True if touching, False otherwise.

        This method **must** be overridden.
        """

    @abstractmethod
    def translate(self, pdisp):
        """
        Perform necessary processing in response to being moved by the mouse/UI.

        :param tuple(int,int) pdisp: Translation vector (x,y) in physical screen
            units.
        :returns: None

        This method **must** be overridden.
        """

    def stroke(self, ppos, pdisp):
        """
        Perform necessary processing in response to click-drag action by the
        mouse/UI.

        :param tuple(int,int) ppos: Physical coordinates of stroke start.
        :param tuple(int,int) pdisp: Translation vector of stroke action in
            physical screen units.
        :returns: None

        This method is intended to be overridden.
        """

    def canStroke(self, ppos):
        """
        Can the object respond to beginning a stroke action at the given
        position.

        :param tuple(int,int) ppos: Physical coordinates of stroke start.
        :rtype: Boolean
        :returns: True if the object can respond, False otherwise.

        This method is intended to be overridden.
        """

    def touchAsset(self, force=False):
        """
        Check to see if an asset needs to be updated it and if so (or forced)
        call the :func:`_updateAsset` method.
        """
        inputs = self._getInputs()
        changed = self._inputsChanged(inputs)
        if changed:
            self._saveInputs(inputs)
        if changed or force:
            self._updateAsset(self._buildAsset())

    @abstractmethod
    def _buildAsset(self):
        pass



"""
Circle object for MathApp applications
"""

from math import sqrt
from ggame.mathapp import MathApp, _MathVisual
from ggame.asset import CircleAsset, PolygonAsset, Color


class Circle(_MathVisual):
    """
    Create a circle on the screen. This is a subclass of
    :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited members are of
    little use and are not shown in the documentation.

    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos** (*tuple(float,float)*) Center point of the circle, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.
        * **radius** [float or Point] Radius of the circle (logical units)
            or a :class:`~ggame.point.Point` on the circle.


    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **style** (*LineStyle*) Valid :class:`~ggame.asset.LineStyle` object
        * **color** (*Color*) Valid :class:`~ggame.color.Color` object`

    Example:

    .. literalinclude:: ../examples/circlecircle.py

    """

    _posinputsdef = ["pos"]
    _nonposinputsdef = ["radius"]
    _defaultcolor = Color(0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(
            CircleAsset(0, self._defaultstyle, self._defaultcolor), *args, **kwargs
        )
        self.touchAsset()
        self.fxcenter = self.fycenter = 0.5

    def _buildAsset(self):
        pcenter = self._spposinputs.pos
        try:
            pradius = (
                MathApp.distance(self._posinputs.pos(), self._nposinputs.radius())
                * MathApp.scale
            )
        except (AttributeError, TypeError):
            # pylint: disable=no-member
            pradius = self._nposinputs.radius() * MathApp.scale
            # pylint: enable=no-member
        style = self._stdinputs.style()
        fill = self._stdinputs.color()
        ymax = pcenter[1] + pradius
        ymin = pcenter[1] - pradius
        xmax = pcenter[0] + pradius
        xmin = pcenter[0] - pradius
        try:
            if ymin > MathApp.height or ymax < 0 or xmax < 0 or xmin > MathApp.width:
                return CircleAsset(pradius, style, fill)
            if pradius > 2 * MathApp.width:
                # here begins unpleasant hack to overcome crappy circles
                poly = self._buildPolygon(pcenter, pradius)
                if poly:
                    passet = PolygonAsset(poly, style, fill)
                    return passet
        except AttributeError:
            return CircleAsset(pradius, style, fill)
        return CircleAsset(pradius, style, fill)

    def _buildPolygon(self, pcenter, pradius):
        """
        pcenter is in screen relative coordinates.
        returns a coordinate list in circle relative coordinates
        """
        xcepts = [
            self._findIntercepts(pcenter, pradius, 0, 0, 0, MathApp.height),
            self._findIntercepts(pcenter, pradius, 0, 0, MathApp.width, 0),
            self._findIntercepts(
                pcenter, pradius, MathApp.width, 0, MathApp.width, MathApp.height
            ),
            self._findIntercepts(
                pcenter, pradius, 0, MathApp.height, MathApp.width, MathApp.height
            ),
        ]
        ilist = []
        for x in xcepts:
            if x and len(x) < 2:
                ilist.extend(x)
        # ilist is a list of boundary intercepts that are screen-relative
        if len(ilist) > 1:
            xrange = ilist[-1][0] - ilist[0][0]
            yrange = ilist[-1][1] - ilist[0][1]
            numpoints = 20
            inx = 0
            for i in range(numpoints):
                icepts = self._findIntercepts(
                    pcenter,
                    pradius,
                    pcenter[0],
                    pcenter[1],
                    ilist[0][0] + xrange * (i + 1) / (numpoints + 1),
                    ilist[0][1] + yrange * (i + 1) / (numpoints + 1),
                )
                if icepts:
                    ilist.insert(inx + 1, icepts[0])
                    inx = inx + 1
            self._addBoundaryVertices(ilist)
            ilist.append(ilist[0])
            ilist = [(i[0] - pcenter[0], i[1] - pcenter[1]) for i in ilist]
        return ilist

    def _addBoundaryVertices(self, plist):
        """
        Sides 0=top, 1=right, 2=bottom, 3=left
        """
        # figure out rotation in point sequence
        cw = 0
        try:
            rtst = plist[0:3] + [plist[0]]
            for p in range(3):
                cw = cw + (rtst[p + 1][0] - rtst[p][0]) * (rtst[p + 1][1] + rtst[p][1])
        except IndexError:
            # print(plist)
            return
        cw = self._sgn(cw)
        cw = 1 if cw < 0 else 0
        vertices = (
            (-100, -100),
            (MathApp.width + 100, -100),
            (MathApp.width + 100, MathApp.height + 100),
            (-100, MathApp.height + 100),
        )
        nextvertex = [
            (vertices[0], vertices[1]),
            (vertices[1], vertices[2]),
            (vertices[2], vertices[3]),
            (vertices[3], vertices[0]),
        ]
        nextsides = [(3, 1), (0, 2), (1, 3), (2, 0)]
        edges = ((None, 0), (MathApp.width, None), (None, MathApp.height), (0, None))
        endside = startside = None
        for side in range(4):
            if endside is None and (
                edges[side][0] == round(plist[-1][0])
                or edges[side][1] == round(plist[-1][1])
            ):
                endside = side
            if startside is None and (
                edges[side][0] == round(plist[0][0])
                or edges[side][1] == round(plist[0][1])
            ):
                startside = side
        iterations = 0
        while startside != endside:
            iterations = iterations + 1
            if iterations > 20:
                break
            if (
                endside is not None and startside is not None
            ):  # and endside != startside
                plist.append(nextvertex[endside][cw])
                endside = nextsides[endside][cw]

    @staticmethod
    def _sgn(x):
        return 1 if x >= 0 else -1

    def _findIntercepts(self, c, r, x1, y1, x2, y2):
        """
        c (center) and x and y values are physical, screen relative.
        function returns coordinates in screen relative format
        """

        def getcoords(x, y):
            if 0 <= x <= MathApp.width and 0 <= y <= MathApp.height:
                return [(x, y)]
            return []

        x1n = x1 - c[0]
        x2n = x2 - c[0]
        y1n = y1 - c[1]
        y2n = y2 - c[1]
        dx = x2n - x1n
        dy = y2n - y1n
        dr = sqrt(dx * dx + dy * dy)
        d = x1n * y2n - x2n * y1n
        disc = r * r * dr * dr - d * d
        dr2 = dr * dr
        if disc <= 0:  # less than two solutions
            return []
        sdisc = sqrt(disc)
        x = [
            (d * dy + self._sgn(dy) * dx * sdisc) / dr2 + c[0],
            (d * dy - self._sgn(dy) * dx * sdisc) / dr2 + c[0],
        ]
        y = [
            (-d * dx + abs(dy) * sdisc) / dr2 + c[1],
            (-d * dx - abs(dy) * sdisc) / dr2 + c[1],
        ]
        res = getcoords(x[0], y[0])
        res.extend(getcoords(x[1], y[1]))
        return res

    @property
    def center(self):
        """
        An ordered pair (x,y) or :class:`~ggame.point.Point` that represents
        the (logical) circle center. This attribute is only get-able.
        """
        # pylint: disable=no-member
        return self._posinputs.pos()
        # pylint: enable=no-member

    @property
    def radius(self):
        """
        A **float** that represents the radius of the circle. This attribugte
        is only get-able.
        """
        # pylint: disable=no-member
        return self._nposinputs.radius()
        # pylint: enable=no-member

    def step(self):
        self.touchAsset()

    def physicalPointTouching(self, ppos):
        r = MathApp.distance(self._spposinputs.pos(), ppos)
        # pylint: disable=no-member
        pradius = self._nposinputs.radius() * MathApp.scale
        # pylint: enable=no-member
        style = self._stdinputs.style()
        inner = pradius - style.width / 2
        outer = pradius + style.width / 2
        return inner <= r <= outer

    def translate(self, pdisp):
        pass



"""
The ggame :class:`App` class encapsulates functionality required for
initiating a graphics window in the browser, executing code using Javascript
animate, routing browser UI events (e.g. mouse and keyboard input) to
user Python code, and managing graphics elements (instances of
:class:`Sprite`, for example).
"""

# app.py

import traceback
from ggame.sysdeps import GFX_Window
from ggame.event import MouseEvent, KeyEvent


class App:
    """
    The :class:`App` class is a (typically subclassed) class that encapsulates
    handling of the display system, and processing user events. The :class:`App`
    class also manages lists of all :class:`Sprite` instances in the
    application.

    When subclassing :class:`App` you may elect to instantiate most of your
    sprite objects in the initialization section.

    Processing that must occur on a per-frame basis may be included by
    overriding the :meth:`~App.run` method. This is also an appropriate
    location to call similar 'step' methods for your various customized sprite
    classes.

    Once your application class has been instantiated, begin the frame drawing
    process by calling its :meth:`~App.run` method.

    The :class:`App` class is instantiated either by specifying the desired app
    window size in pixels, as two parameters:::

        myapp = App(640,480)

    or by providing no size parameters at all:::

        myapp = App()

    in which case, the full browser window size is used.

    NOTE: Only **one** instance of an :class:`App` class or subclass may be
    instantiated at a time.
    """

    spritelist = []
    """
    List of all sprites currently active in the application.
    """
    _eventdict = {}
    _spritesdict = {}
    _spritesadded = False
    win = None

    def __init__(self, *args):
        if App.win is None and (not args or len(args) == 2):
            x = y = 0
            if len(args) == 2:
                x = args[0]
                y = args[1]
            App.win = GFX_Window(x, y, type(self).destroy)
            App.width = App.win.width
            App.height = App.win.height
            # Add existing sprites to the window
            if not App._spritesadded and App.spritelist:
                App._spritesadded = True
                for sprite in App.spritelist:
                    App.win.add(sprite.gfx)
            App.win.bind(KeyEvent.keydown, type(self)._keyEvent)
            App.win.bind(KeyEvent.keyup, type(self)._keyEvent)
            App.win.bind(KeyEvent.keypress, type(self)._keyEvent)
            App.win.bind(MouseEvent.mousewheel, type(self)._mouseEvent)
            App.win.bind(MouseEvent.mousemove, type(self)._mouseEvent)
            App.win.bind(MouseEvent.mousedown, type(self)._mouseEvent)
            App.win.bind(MouseEvent.mouseup, type(self)._mouseEvent)
            App.win.bind(MouseEvent.click, type(self)._mouseEvent)
            App.win.bind(MouseEvent.dblclick, type(self)._mouseEvent)
        self.userfunc = None

    @classmethod
    def _routeEvent(cls, event, evtlist):
        for callback in reversed(evtlist):
            if not event.consumed:
                try:
                    callback(event)
                except BaseException:
                    traceback.print_exc()
                    raise

    @classmethod
    def _keyEvent(cls, hwevent):
        evtlist = App._eventdict.get(
            (hwevent.type, KeyEvent.keys.get(hwevent.keyCode, 0)), []
        )
        evtlist.extend(App._eventdict.get((hwevent.type, "*"), []))
        if evtlist:
            evt = KeyEvent(hwevent)
            cls._routeEvent(evt, evtlist)
        return False

    @classmethod
    def _mouseEvent(cls, hwevent):
        evtlist = App._eventdict.get(hwevent.type, [])
        if evtlist:
            evt = MouseEvent(cls, hwevent)
            cls._routeEvent(evt, evtlist)
        return False

    @classmethod
    def add(cls, obj):
        """
        Add a sprite object to the global lists.

        :param Sprite obj: The sprite reference to add.
        :returns: None
        """
        if App.win is not None:
            App.win.add(obj.gfx)
        App.spritelist.append(obj)
        # find out if sprite type is in dictionary without triggering pylint
        if not App._spritesdict.get(type(obj), False):
            App._spritesdict[type(obj)] = []
        App._spritesdict[type(obj)].append(obj)

    @classmethod
    def remove(cls, obj):
        """
        Remove a sprite object from the global lists.

        :param Sprite obj: The sprite reference to remove.
        :returns: None
        """
        App.spritelist.remove(obj)
        # remove from underlying layer only if existed in ours
        if App.win is not None:
            App.win.remove(obj.gfx)
        App._spritesdict[type(obj)].remove(obj)

    def _animate(self, _dummy):
        if App.win:
            try:
                if self.userfunc:
                    self.userfunc()
                else:
                    self.step()
            except BaseException:
                traceback.print_exc()
                raise
            App.win.animate(self._animate)

    @classmethod
    def destroy(cls):
        """
        This will close the display window/tab, remove all references to
        sprites and place the `App` class in a state in which a new
        application could be instantiated.
        """
        if App.win:
            App.win.unbind(KeyEvent.keydown)
            App.win.unbind(KeyEvent.keyup)
            App.win.unbind(KeyEvent.keypress)
            App.win.unbind(MouseEvent.mousewheel)
            App.win.unbind(MouseEvent.mousemove)
            App.win.unbind(MouseEvent.mousedown)
            App.win.unbind(MouseEvent.mouseup)
            App.win.unbind(MouseEvent.click)
            App.win.unbind(MouseEvent.dblclick)
        for s in list(App.spritelist):
            s.destroy()
        App.win.destroy()
        App.win = None
        App.spritelist = []
        App._spritesdict = {}
        App._eventdict = {}
        App._spritesadded = False

    @classmethod
    def listenKeyEvent(cls, eventtype, key, callback):
        """
        Register to receive keyboard events.

        :param str eventtype:  The type of key event to
            receive (value is one of: `'keydown'`, `'keyup'` or `'keypress'`).

        :param str key:  Identify the keyboard key (e.g. `'space'`,
            `'left arrow'`, etc.) to receive events for.

        :param function callback:  The function or method that will be
            called with the :class:`~ggame.event.KeyEvent` object when the
            event occurs.

        :returns: Nothing

        See the source for :class:`~ggame.event.KeyEvent` for a list of key
        names
        to use with the `key` paramter.
        """
        evtlist = App._eventdict.get((eventtype, key), [])
        if callback not in evtlist:
            evtlist.append(callback)
        App._eventdict[(eventtype, key)] = evtlist

    @classmethod
    def listenMouseEvent(cls, eventtype, callback):
        """
        Register to receive mouse events.

        :param str eventtype: The type of mouse event to
            receive (value is one of: `'mousemove'`, `'mousedown'`, `'mouseup'`,
            `'click'`, `'dblclick'` or `'mousewheel'`).

        :param function callback: The function or method that will be
            called with the :class:`ggame.event.MouseEvent` object when the
            event occurs.

        :returns: Nothing
        """
        evtlist = App._eventdict.get(eventtype, [])
        if callback not in evtlist:
            evtlist.append(callback)
        App._eventdict[eventtype] = evtlist

    @classmethod
    def unlistenKeyEvent(cls, eventtype, key, callback):
        """
        Use this method to remove a registration to receive a particular
        keyboard event. Arguments must exactly match those used when
        registering for the event.

        :param str eventtype:  The type of key event to stop
            receiving (value is one of: `'keydown'`, `'keyup'` or `'keypress'`).

        :param str key:  The keyboard key (e.g. `'space'`,
            `'left arrow'`, etc.) to stop receiving events for.

        :param function callback:  The function or method that will no longer
            be called with the :class:`~ggame.event.KeyEvent` object when the
            event occurs.

        :returns: Nothing

        See the source for :class:`~ggame.event.KeyEvent` for a list of key
        names to use with the `key` paramter.

        """
        App._eventdict[(eventtype, key)].remove(callback)

    @classmethod
    def unlistenMouseEvent(cls, eventtype, callback):
        """
        Use this method to remove a registration to receive a particular
        mouse event. Arguments must exactly match those used when
        registering for the event.

        :param str eventtype: The type of mouse event to stop receiving events
            for (value is one of: `'mousemove'`, `'mousedown'`, `'mouseup'`,
            `'click'`, `'dblclick'` or `'mousewheel'`).

        :param function callback: The function or method that will no longer be
            called with the :class:`ggame.event.MouseEvent` object when the
            event occurs.

        :returns: Nothing
        """
        App._eventdict[eventtype].remove(callback)

    @classmethod
    def getSpritesbyClass(cls, sclass):
        """
        Returns a list of all active sprites of a given class.

        :param class sclass: A class name (e.g. 'Sprite') or subclass.

        :returns: A (potentially empty) list of sprite references.
        """
        return App._spritesdict.get(sclass, [])[:]

    def step(self):
        """
        The :meth:`~App.step` method is called once per animation frame.
        Override this method in your own subclass of :class:`App` to perform
        periodic calculations, such as checking for sprite collisions, or
        calling 'step' functions in your own customized sprite classes.

        The base class :meth:`~App.step` method is empty and is intended to be
        overriden.

        :returns: Nothing

        """

    def run(self, userfunc=None):
        """
        Calling the :meth:`~App.run` method begins the animation process
        whereby the :meth:`~App.step` method is called once per animation frame.

        :param function userfunc: Any function or method which shall be
            called once per animation frame.

        :returns: Nothing
        """
        self.userfunc = userfunc
        App.win.animate(self._animate)



"""
MathApp GUI input pushbutton and toggle controls.
"""

from ggame.point import ImagePoint
from ggame.asset import Frame


class InputImageButton(ImagePoint):
    """
    :class:`~ggame.point.ImagePoint` object that uses an image as its on-screen and
        activates a callback function when pressed.

    :Required Arguments:

    :param str url: Location of an image file (png, jpg)
    :param function callback: Reference to a function to execute, passing this
        button object.

    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos** (*tuple(float,float)*) Position in physical or logical units.

    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' (default) or 'physical'
        * **frame** (*Frame*) The sub-frame location of image within the image file
        * **qty** (*int*) The number of sub-frames, when used as a sprite sheet
        * **direction** (*str*) One of 'horizontal' (default) or 'vertical'
        * **margin** (*int*) Pixels between sub-frames if sprite sheet

    Example:

    .. literalinclude:: ../examples/inputpointinputimagebutton.py

    """

    def __init__(self, url, callback, *args, **kwargs):
        super().__init__(url, *args, **kwargs)
        self.center = (0, 0)
        self._callback = callback
        self.selectable = True
        self.firstImage()
        self.mousewasdown = self.mouseisdown

    def select(self):
        super().select()
        if self._callback:
            self._callback(self)
        self.unselect()

    def __call__(self):
        # code for controlling the button image state only works if the
        # button state is being monitored!
        if self.mouseisdown != self.mousewasdown:
            if self.mouseisdown:
                self.nextImage()
            else:
                self.firstImage()
            self.mousewasdown = self.mouseisdown
        return self.mouseisdown


class InputImageToggle(ImagePoint):
    """
    Spritesheet-based input control that toggles through a set of states with each
    mouseclick.

    :Required Inputs:

    :param str url: Location of image file consisting of a multi-image sprite sheet.
    :param list statelist: List of values to correspond with toggle states.
    :param int initindex: Index to initial toggle state.
    :param (float,float) pos: Position in physical or logical units. May be a
        :class:`ggame.point.Point` instance, a literal (x,y) pair, or a function
        that returns an (x,y) pair.

    :param \\**kwargs:
        See below

    :Optional Keyword Arguments:

    * **positioning** (*str*) One of 'logical' (default) or 'physical'
    * **frame** (*:class:`ggame.asset.Frame`*) Sub-frame location of sub-image within
        the main image.
    * **direction** (*str*) One of 'horizontal' (default) or 'vertical'.
    * **margin** (*int*) Number of pixels between sub-frames if sprite sheet.

    """

    def __init__(self, url, statelist, initindex, *args, **kwargs):
        self.statelist = statelist
        kwargs.setdefault("qty", len(statelist))
        super().__init__(url, *args, **kwargs)
        self.center = (0, 0)
        self.selectable = True
        self.togglestate = initindex
        self.setImage(self.togglestate)

    def select(self):
        super().select()
        self.togglestate += 1
        if self.togglestate == len(self.statelist):
            self.togglestate = 0
        self.setImage(self.togglestate)
        self.unselect()

    def __call__(self):
        return self.statelist[self.togglestate]


class MetalToggle(InputImageToggle):
    """
    Metal toggle input control that toggles through two states with each
    mouseclick.

    :Required Inputs:

    :param int initindex: Index to initial toggle state (0 or 1).
    :param (float,float) pos: Position in physical or logical units. May be a
        :class:`ggame.point.Point` instance, a literal (x,y) pair, or a function
        that returns an (x,y) pair.

    Example:

    .. literalinclude:: ../examples/indicatorledindicator.py

    """

    def __init__(self, initindex, *args, **kwargs):
        kwargs.setdefault("frame", Frame(0, 0, 110, 150))
        super().__init__(
            self.getImagePath("toggle-up-down.png"),
            [True, False],
            initindex,
            *args,
            **kwargs
        )
        self.scale = 0.4


class GlassButton(InputImageButton):
    """
    Glass button input control that triggers a callback when clicked.

    :Required Inputs:

    :param function callback: Reference to a function to execute, passing this button
        object as an argument.
    :param (float,float) pos: Position in physical or logical units. May be a
        :class:`ggame.point.Point` instance, a literal (x,y) pair, or a function
        that returns an (x,y) pair.

    Example:

    .. literalinclude:: ../examples/inputpointglassbutton.py

    """

    def __init__(self, callback, *args, **kwargs):
        kwargs.setdefault("frame", Frame(0, 0, 100, 100))
        kwargs.setdefault("qty", 2)
        super().__init__(
            self.getImagePath("button-round.png"), callback, *args, **kwargs
        )
        self.scale = 0.3



# pylint: skip-file


def module_exists(module_name):
    try:
        __import__(module_name)
    except ImportError:
        return False
    else:
        return True


# PYTHON 3 and PYGAME DEPENDENCIES
if module_exists("pygame"):
    import pygame

    class _body(object):
        def __init__(self):
            self.events = {}

        def appendChild(self, obj):
            self.child = obj

        def bind(self, evt, action):
            self.events[evt] = action
            print("Binding {} to {}".format(evt, action))

    class _document(object):
        def __init__(self):
            self.body = _body()

    class _window(object):
        def __init__(self):
            self.document = _document()
            self.animatex = 0

        def open(self, s1, s2):
            return self

        def requestAnimationFrame(self, target):
            if self.animatex < 10:
                self.animatex += 1
                target("dummy")
                print("Animation frame")

    class _Container(object):
        def __init__(self):
            pass

        def destroy(self):
            pass

    class _Renderer(object):
        def __init__(self, x, y, argsdict):
            self.x = x
            self.y = y
            self.argsdict = argsdict
            self.view = "view"
            print("Rendering created with {}x{} area".format(x, y))

        def render(self, stage):
            pass

    class _GFX(object):
        def __init__(self):
            self.Container = _Container
            self.autoDetectRenderer = _Renderer

    window = _window()

    GFX = _GFX()

    # document = object()

    def JSConstructor(cls):
        return cls

    def JSObject(obj):
        return obj

    class _GFX_Rectangle(pygame.Rect):
        pass

    GFX_Rectangle = _GFX_Rectangle

    class _Texture(object):
        def __init__(self, img="", crossdomain=False):
            self.name = img
            if not img == "":
                self.img = pygame.image.load(img)  # pygame surface
                self.basewidth = self.img.get_width()
                self.baseheight = self.img.get_height()
                self.width = self.basewidth
                self.height = self.baseheight
                print(
                    "Texture from image {}, {}x{} pixels".format(
                        img, self.basewidth, self.baseheight
                    )
                )
                self.baserect = _GFX_Rectangle(0, 0, self.basewidth, self.baseheight)
                self.framerect = self.baserect

        @classmethod
        def fromTexture(cls, texture, frame):
            inst = cls()
            inst.img = pygame.Surface((frame.width, frame.height))
            inst.img.blit(texture.img, (0, 0), frame)
            inst.name = texture.name
            inst.basewidth = texture.basewidth
            inst.baseheight = texture.baseheight
            inst.baserect = texture.baserect
            inst.framerect = frame
            inst.width = frame.width
            inst.height = frame.height
            print(
                "Texture from base texture {}, {}x{} subframe {}x{}".format(
                    inst.name,
                    inst.basewidth,
                    inst.baseheight,
                    inst.framerect.width,
                    inst.framerect.height,
                )
            )
            return inst

        def destroy(self):
            try:
                self.img.close()
                print("Destroying an image")
            except:
                print("Destroying a non-image")

    GFX_Texture = _Texture.fromTexture

    GFX_Texture_fromImage = _Texture

    class vector(object):
        def __init__(self, x, y):
            self.x = x
            self.y = y

        def __getitem__(self, key):
            if key == 0:
                return self.x
            elif key == 1:
                return self.y
            else:
                raise KeyError

        def __setitem(self, key, value):
            if key == 0:
                self.x = value
            elif key == 1:
                self.y = value
            else:
                raise KeyError

    class GFX_Sprite(object):
        def __init__(self, texture):
            self.basetexture = texture
            self.texture = self.basetexture
            self.visible = True
            self.pos = vector(0, 0)
            self.anch = vector(0, 0)
            self.scal = vector(1.0, 1.0)
            self.width = texture.width
            self.height = texture.height
            self.rotation = 0.0

        @property
        def position(self):
            return self.pos

        @position.setter
        def position(self, value):
            self.pos.x = value[0]
            self.pos.y = value[1]

        @property
        def anchor(self):
            return self.anch

        @anchor.setter
        def anchor(self, value):
            self.anch.x = value[0]
            self.anch.y = value[1]

        @property
        def scale(self):
            return self.scal

        @scale.setter
        def scale(self, value):
            self.scal.x = value[0]
            self.scal.y = value[1]

        def destroy(self):
            pass

    class _GFX_Graphics(object):
        def __init__(self):
            self.clear()

        def clear(self):
            self.cleared = True
            self.visible = True
            self.lwidth = None
            self.color = None
            self.alpha = None
            self.fillcolor = None
            self.fillalpha = None
            self.x = None
            self.y = None
            self.rwidth = None
            self.rheight = None
            self.radius = None
            self.ehw = None
            self.ehh = None
            self.xto = None
            self.yto = None
            self.jpath = None
            self.width = None
            self.height = None
            self.position = vector(0, 0)

        def destroy(self):
            self.clear()

        def clone(self):
            clone = type(self)()
            clone.cleared = self.cleared
            clone.visible = self.visible
            clone.lwidth = self.lwidth
            clone.color = self.color
            clone.alpha = self.alpha
            clone.fillalpha = self.fillalpha
            clone.fillcolor = self.fillcolor
            clone.x = self.x
            clone.y = self.y
            clone.rwidth = self.rwidth
            clone.rheight = self.rheight
            clone.radius = self.radius
            clone.ehw = self.ehw
            clone.ehh = self.ehh
            clone.xto = self.xto
            clone.yto = self.yto
            clone.jpath = self.jpath
            clone.width = self.width
            clone.height = self.height
            clone.position = self.position
            return clone

        def lineStyle(self, width, color, alpha):
            self.width = width
            self.color = color
            self.alpha = alpha

        def beginFill(self, color, alpha):
            self.fillcolor = color
            self.fillalpha = alpha

        def drawRect(self, x, y, w, h):
            self.x = x
            self.y = y
            self.position = vector(x, y)
            self.rwidth = w
            self.rheight = h
            self.width = w
            self.height = h
            self.cleared = False
            print("Rectangle {}x{} at {},{}".format(w, h, x, y))
            return self

        def drawCircle(self, x, y, radius):
            self.x = x
            self.y = y
            self.position = vector(x, y)
            self.radius = radius
            self.cleared = False
            self.width = radius * 2
            self.height = radius * 2
            print("Circle, radius {} at {},{}".format(radius, x, y))
            return self

        def drawEllipse(self, x, y, hw, hh):
            self.x = x
            self.y = y
            self.position = vector(x, y)
            self.ehw = hw
            self.ehh = hh
            self.width = hw * 2
            self.height = hh * 2
            self.cleared = False
            print("Ellipse, {}x{} at {},{}".format(hw, hh, x, y))
            return self

        def drawPolygon(self, jpath):
            self.jpath = jpath
            self.cleared = False
            self.position = vector(jpath[0], jpath[1])
            x = []
            y = []
            for i in range(0, len(jpath) - 1, 2):
                x.append(jpath[i])
                y.append(jpath[i + 1])
            self.width = max(x) - min(x)
            self.height = max(y) - min(y)
            print("Polygon")
            return self

        def moveTo(self, x, y):
            self.x = x
            self.y = y
            self.position = vector(x, y)
            return self

        def lineTo(self, x, y):
            self.xto = x
            self.yto = y
            self.width = abs(x)
            self.height = abs(y)
            self.cleared = False
            print("Line from {},{} to {},{}".format(self.x, self.y, x, y))
            return self

    class _GFX_Text(object):
        def __init__(self, text, styledict):
            self.text = text
            self.styledict = styledict
            self.alpha = None
            self.visible = None
            self.width = 99
            self.height = 99
            self.position = vector(0, 0)
            print("Text: {} in {}".format(text, styledict["font"]))

        def clone(self):
            clone = type(self)(self.text, self.styledict)
            return clone

        def destroy(self):
            self.text = ""

    GFX_Text = _GFX_Text

    _globalGraphics = _GFX_Graphics()

    GFX_Graphics = _globalGraphics

    def GFX_DetectRenderer():
        pass

    class _SND_all(object):
        def __init__(self):
            pass

        def stop(self):
            print("Stopping all sounds")

    class _SND(object):
        def __init__(self):
            self.all = _SND_all

    SND = _SND()

    class _SND_Sound(object):
        def __init__(self, url):
            self.url = url
            print("Creating sound object {}".format(url))

        def load(self):
            pass

        def play(self):
            print("Playing sound object {}".format(self.url))

    SND_Sound = _SND_Sound

    class HwEvent(object):
        evtmap = {
            2: "keydown",
            3: "keyup",
            4: "mousemove",
            5: "mousedown",
            6: "mouseup",
        }
        keymap = {
            304: 16,
            303: 16,
            306: 17,
            308: 18,
            301: 20,
            276: 37,
            273: 38,
            275: 39,
            274: 40,
            97: 65,
            98: 66,
            99: 67,
            100: 68,
            101: 69,
            102: 70,
            103: 71,
            104: 72,
            105: 73,
            106: 74,
            107: 75,
            108: 76,
            109: 77,
            110: 78,
            111: 79,
            112: 80,
            113: 81,
            114: 82,
            115: 83,
            116: 84,
            117: 85,
            118: 86,
            119: 87,
            120: 88,
            121: 89,
            122: 90,
            282: 112,
            283: 113,
            284: 114,
            285: 115,
            286: 116,
            287: 117,
            288: 118,
            289: 119,
            290: 120,
            291: 121,
            292: 122,
            293: 123,
            59: 186,
            61: 187,
            44: 188,
            46: 190,
            45: 189,
            47: 191,
            96: 192,
            92: 220,
            91: 219,
            93: 221,
            39: 222,
        }

        def __init__(self, pevent):
            self.type = HwEvent.evtmap.get(pevent.type, None)
            if self.type in ["keydown", "keyup"]:
                self.keyCode = HwEvent.keymap.get(pevent.key, pevent.key)
            elif self.type in ["mousemove", "mousedown", "mouseup"]:
                self.wheelDelta = 0
                if self.type != "mousemove" and pevent.button == 5:
                    if self.type == "mousedown":
                        self.wheelDelta = 1
                    else:
                        self.wheelDelta = -1
                self.clientX = pevent.pos[0]
                self.clientY = pevent.pos[1]

    class GFX_Window(object):
        def __init__(self, width, height, onclose):
            pygame.init()
            self._w = pygame.display.set_mode((width, height))
            self.clock = pygame.time.Clock()
            self.sprites = []
            self.animatestarted = False
            self.bindings = {}
            self.onclose = onclose
            self.stop = False
            # self._w = window.open("", "")
            # self._stage = JSConstructor(GFX.Container)()
            # self.renderer = GFX.autoDetectRenderer(width, height, {'transparent':True})
            # self._w.document.body.appendChild(self.renderer.view)
            # self._w.onunload = onclose

        def bind(self, evtspec, callback):
            self.bindings[evtspec] = callback

        def add(self, obj):
            self.sprites.append(obj)
            # self._stage.addChild(obj)

        def remove(self, obj):
            self.sprites.remove(obj)
            # self._stage.removeChild(obj)

        def animate(self, stepcallback):
            # do stuff required to display
            self._w.fill(pygame.Color("white"))
            for s in self.sprites:
                self._w.blit(s.texture.img, (s.pos.x, s.pos.y))
            pygame.display.flip()
            events = pygame.event.get()
            for event in events:
                hwevent = HwEvent(event)
                if hwevent.type != None:
                    self.bindings[hwevent.type](hwevent)
                if event.type == 12:
                    print("Close!")
                    self.onclose()
                    self.destroy()
                    self.stop = True
            if not self.animatestarted:
                self.animatestarted = True
                while not self.stop:
                    self.clock.tick_busy_loop(30)
                    stepcallback(0)
            # self.renderer.render(self._stage)
            # self._w.requestAnimationFrame(stepcallback)

        def destroy(self):
            pass
            # SND.all().stop()
            # self._stage.destroy()



"""
MathApp input class for accepting user numeric input with a "slider" control.
"""

from ggame.mathapp import MathApp, _MathVisual
from ggame.asset import RectangleAsset, LineStyle, Color
from ggame.sprite import Sprite


class Slider(_MathVisual):
    """
    Create a 'slider' style numeric input control on the screen. This is a
    subclass of :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited
    members are of little use and are not shown in the documentation.


    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos**  (*tuple(float,float)*) Screen position of the slider, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.
        * **minval** (*float*) The minimum value of the slider
        * **maxval** (*float*) The maximum value of the slider
        * **initial** (*float*) The initial value of the slider


    :Optional Keyword Arguments:
        * **steps** (*int*) Number of steps between minval and maxval (default 50)
        * **leftkey** (*str*) Name of a keyboard key that will make the
            slider control step down (move left) (default None). See
            :class:`~ggame.event.KeyEvent` for a list of names.
        * **rightkey** (*str*) Name of a keyboard key that will make the slider
            control step up (move right) (default None)
        * **centerkey** (*str*) Name of a keyboard key that will make the slider
            move to its center position (default None)
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **size** (*int*) Width of the slider (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object
        * **style** (*LineStyle*) Valid :class:`~ggame.asset.LineStyle` object

    Example:

    .. literalinclude:: ../examples/sliderslider.py

    """

    _posinputsdef = ["pos"]
    _nonposinputsdef = ["minval", "maxval", "initial"]

    def __init__(self, *args, **kwargs):
        super().__init__(RectangleAsset(1, 1), *args, **kwargs)
        self._val = self._nposinputs.initial()  # pylint: disable=no-member
        self._steps = kwargs.get("steps", 50)
        self._step = (
            self._nposinputs.maxval()  # pylint: disable=no-member
            - self._nposinputs.minval()  # pylint: disable=no-member
        ) / self._steps
        self._leftctrl = kwargs.get("leftkey", None)
        self._rightctrl = kwargs.get("rightkey", None)
        self._centerctrl = kwargs.get("centerkey", None)
        self.selectable = True  # must be after super init!
        self.strokable = True  # this enables grabbing/slideing the thumb
        self.thumbcaptured = False
        self._thumbwidth = max(self._stdinputs.width() / 40, 1)
        self._thumb = Sprite(
            RectangleAsset(
                self._thumbwidth,
                self._stdinputs.size() - 2,
                LineStyle(1, self._stdinputs.color()),
                self._stdinputs.color(),
            ),
            self._thumbXY(),
        )
        self.touchAsset()
        if self._leftctrl:
            MathApp.listenKeyEvent("keydown", self._leftctrl, self._moveLeft)
        if self._rightctrl:
            MathApp.listenKeyEvent("keydown", self._rightctrl, self._moveRight)
        if self._centerctrl:
            MathApp.listenKeyEvent("keydown", self._centerctrl, self._moveCenter)

    def _thumbXY(self):
        minval = self._nposinputs.minval()  # pylint: disable=no-member
        maxval = self._nposinputs.maxval()  # pylint: disable=no-member
        return (
            self._spposinputs.pos[0]
            + (self._val - minval)
            * (self._sstdinputs.width - self._thumbwidth)
            / (maxval - minval),
            self._spposinputs.pos[1] + 1,
        )

    def __call__(self):
        return self._val

    @property
    def value(self):
        """
        Report value of the slider. Attribute is get-able and set-able.
        """
        return self._val

    @value.setter
    def value(self, val):
        self._setval(val)

    def _buildAsset(self):
        self._setThumb()
        return RectangleAsset(
            self._stdinputs.width(),
            self._stdinputs.size(),
            line=self._stdinputs.style(),
            fill=Color(0, 0),
        )

    def _setThumb(self):
        self._thumb.position = self._thumbXY()

    def step(self):
        pass

    def _setval(self, val):
        minval = self._nposinputs.minval()  # pylint: disable=no-member
        maxval = self._nposinputs.maxval()  # pylint: disable=no-member
        if val <= minval:
            self._val = minval
        elif val >= maxval:
            self._val = maxval
        else:
            self._val = (
                round((val - minval) * self._steps / (maxval - minval)) * self._step
                + minval
            )
        self._setThumb()

    def increment(self, step):
        """
        Increment the slider value.

        :param float step: The amount by which the slider control should be adjusted.
        :returns: None
        """

        self._setval(self._val + step)

    def select(self):
        super().select()
        if not self._leftctrl:
            MathApp.listenKeyEvent("keydown", "left arrow", self._moveLeft)
        if not self._rightctrl:
            MathApp.listenKeyEvent("keydown", "right arrow", self._moveRight)
        MathApp.listenMouseEvent("click", self._mouseClick)

    def unselect(self):
        super().unselect()
        try:
            if not self._leftctrl:
                MathApp.unlistenKeyEvent("keydown", "left arrow", self._moveLeft)
            if not self._rightctrl:
                MathApp.unlistenKeyEvent("keydown", "right arrow", self._moveRight)
            MathApp.unlistenMouseEvent("click", self._mouseClick)
        except ValueError:
            pass

    def _mouseClick(self, event):
        if self.physicalPointTouching((event.x, event.y)):
            if event.x > self._thumb.x + self._thumbwidth:
                self._moveRight(event)
            elif event.x < self._thumb.x:
                self._moveLeft(event)

    def _moveLeft(self, _event):
        self.increment(-self._step)

    def _moveRight(self, _event):
        self.increment(self._step)

    def _moveCenter(self, _event):
        self._val = (self._snposinputs.minval + self._snposinputs.maxval) / 2
        self._setThumb()

    def canstroke(self, ppos):
        """
        Function returns true if the given physical position corresponds with a part
        of the slider that can be dragged (i.e. the thumb).

        :param (float,float) ppos: Physical screen coordinates.
        :return: True if the position represents a draggable part of the control.
        :rtype: boolean

        """
        return self.physicalPointTouchingThumb(ppos)

    def stroke(self, ppos, pdisp):
        """
        Function performs the action of stroking or click-dragging on the slider
        control (i.e. dragging the thumb).

        :param (float, float) ppos: Physical screen coordinates.
        :param (float, float) pdisp: Physical displacement vector.
        :return: None

        """
        _ppos = self._spposinputs.pos
        minval = self._snposinputs.minval
        maxval = self._snposinputs.maxval
        xpos = ppos[0] + pdisp[0]
        self.value = (xpos - _ppos[0]) * (
            maxval - minval
        ) / self._sstdinputs.width + minval

    def physicalPointTouching(self, ppos):
        _ppos = self._spposinputs.pos
        return (
            ppos[0] >= _ppos[0]
            and ppos[0] <= _ppos[0] + self._sstdinputs.width
            and ppos[1] >= _ppos[1]
            and ppos[1] <= _ppos[1] + self._sstdinputs.size
        )

    def physicalPointTouchingThumb(self, ppos):
        """
        Determine if a physical screen location is touching the slider "thumb".

        :param (float, float) ppos: Physical screen coordinates.
        :returns: True if touching, False otherwise.
        :rtype: boolean
        """

        thumbpos = self._thumbXY()
        return (
            ppos[0] >= thumbpos[0]
            and ppos[0] <= thumbpos[0] + self._thumbwidth + 2
            and ppos[1] >= thumbpos[1]
            and ppos[1] <= thumbpos[1] + self._sstdinputs.size - 2
        )

    def translate(self, pdisp):
        pass

    def destroy(self):
        self._thumb.destroy()
        super().destroy()



"""
Line objects for MathApp applications
"""

from ggame.mathapp import _MathVisual
from ggame.asset import LineAsset


class LineSegment(_MathVisual):
    """
    Create a line segment on the screen. This is a subclass of
    :class:`~ggame.sprite.Sprite` and
    :class:`~ggame.mathapp._MathVisual` but most of the inherited members are of
    little use and are not shown in the documentation.


    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos** (*tuple(float,float)*) Starting point of the segment, which may
            be a literal tuple of floats, or a reference to any object or
            function that returns or evaluates to a tuple of floats.
        * **end** (*tuple(float,float)*) Ending point of the segment (see above)


    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' or 'physical'
        * **style** (*LineStyle*) Valid :class:`~ggame.asset.LineStyle` object

    Example::

        from ggame.point import Point
        from ggame.line import LineSegment
        from ggame.mathapp import MathApp

        p1 = Point((2,1))
        ls = LineSegment(p1, Point((1,1)))

        MathApp().run()
    """

    _posinputsdef = ["pos", "end"]

    def __init__(self, *args, **kwargs):
        super().__init__(LineAsset(0, 0, self._defaultstyle), *args, **kwargs)
        self.touchAsset()

    def _buildAsset(self):
        # pylint: disable=no-member
        start = self._pposinputs.pos
        end = self._pposinputs.end
        # pylint: enable=no-member
        self.position = start
        return LineAsset(end[0] - start[0], end[1] - start[1], self._stdinputs.style())

    def physicalPointTouching(self, ppos):
        """
        This method always returns False.
        """
        return False

    def translate(self, pdisp):
        """
        This method is not implemented.
        """



"""
MathApp indicator classes for displaying two-state or boolean values.
"""

from ggame.mathapp import _MathVisual
from ggame.asset import Frame, ImageAsset


class ImageIndicator(_MathVisual):
    """
    Use a sprite sheet image to indicate integer or boolean values.

    Required Inputs

    :param str url: Location of image file consisting of a multi-image sprite sheet.
    :param (float,float) pos: Position in physical or logical units. May be a
        :class:`ggame.point.Point` instance, a literal (x,y) pair, or a function
        that returns an (x,y) pair.
    :param int value: The state of the indicator. May be a function that returns
        a suitable (integer) value.

    :param \\**kwargs:
        See below


    :Optional Keyword Arguments:

    * **positioning** (*str*) One of 'logical' (default) or 'physical'
    * **frame** (*:class:`ggame.asset.Frame`*) Sub-frame location of sub-image within
        the main image.
    * **qty** (*int*) The number of sub-frames, when used as sprite sheet.
    * **direction** (*str*) One of 'horizontal' (default) or 'vertical'.
    * **margin** (*int*) Number of pixels between sub-frames if sprite sheet.

    Example:

    .. literalinclude:: ../examples/indicatorimageindicator.py


    """

    _posinputsdef = ["pos"]
    _nonposinputsdef = ["value"]

    def __init__(self, url, *args, **kwargs):
        kwargs.setdefault("frame", None)
        kwargs.setdefault("qty", 1)
        kwargs.setdefault("direction", "horizontal")
        kwargs.setdefault("margin", 0)
        super().__init__(
            ImageAsset(
                url,
                kwargs["frame"],
                kwargs["qty"],
                kwargs["direction"],
                kwargs["margin"],
            ),
            *args,
            **kwargs
        )
        self.center = (0, 0)

    def _buildAsset(self):
        # pylint: disable=no-member
        inval = self._nposinputs.value()
        # pylint: enable=no-member
        if inval:
            self.setImage(1)
        elif not inval:
            self.setImage(0)
        else:
            self.setImage(inval)
        return self.asset

    def physicalPointTouching(self, ppos):
        self.setExtents()  # ensure xmin, xmax are correct
        x, y = ppos
        return self.xmin <= x <= self.xmax and self.ymin <= y <= self.ymax

    def translate(self, pdisp):
        pass


class LEDIndicator(ImageIndicator):
    """
    Subclass of ImageIndicator using a red LED on/off image.

    Required Inputs

    :param (float,float) pos: Position in physical or logical units. May be a
        :class:`ggame.point.Point` instance, a literal (x,y) pair, or a function
        that returns an (x,y) pair.
    :param int value: The state of the indicator. May be a function that returns
        a suitable (integer) value.

    :param \\**kwargs:
        See below

    :Optional Keyword Arguments:

        * **positioning** (*str*) One of 'logical' (default) or 'physical'

    Example:

    .. literalinclude:: ../examples/indicatorledindicator.py

    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("frame", Frame(0, 0, 600, 600))
        kwargs.setdefault("qty", 2)
        super().__init__(self.getImagePath("red-led-off-on.png"), *args, **kwargs)
        self.scale = 0.05



"""
Ggame package setup for pypi
"""

import setuptools
from ggame.__version__ import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ggame",
    include_package_data=True,
    version=VERSION,
    author="Eric Dennison",
    author_email="ericd@netdenizen.com",
    description="A lightweight sprite-based game and graphics framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BrythonServer/ggame",
    packages=setuptools.find_packages(exclude=["examples*"]),
    install_requires=("python>=3.11"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
MathApp classes for representing geometric points
"""
from abc import ABCMeta, abstractmethod
from ggame.mathapp import _MathVisual, MathApp
from ggame.asset import Color, LineStyle, CircleAsset, ImageAsset


class _Point(_MathVisual, metaclass=ABCMeta):
    """
    Abstract base class for all point classes.

    :param Asset asset: A valid ggame Asset object
    :param tuple(float,float) pos: The position (physical or logical)
    """

    _posinputsdef = ["pos"]
    _nonposinputsdef = []

    def __init__(self, asset, *args, **kwargs):
        super().__init__(asset, *args, **kwargs)
        self.touchAsset()
        self.center = (0.5, 0.5)

    def __call__(self):
        return self._posinputs.pos()

    def step(self):
        """
        Perform periodic processing.
        """
        self.touchAsset()

    def physicalPointTouching(self, ppos):
        """
        Determine if a physical point is considered to be touching this point.

        :param tuple(int,int) ppos: Physical screen coordinates.
        :rtype: boolean
        :returns: True if touching, False otherwise.
        """
        return (
            MathApp.distance(ppos, self._pposinputs.pos)  # pylint: disable=no-member
            < self._sstdinputs.size
        )

    def translate(self, pdisp):
        """
        Perform necessary processing in response to being moved by the mouse/UI.

        :param tuple(int,int) pdisp: Translation vector (x,y) in physical screen
            units.
        :returns: None
        """
        ldisp = MathApp.translatePhysicalToLogical(pdisp)
        pos = self._posinputs.pos()
        self._posinputs = self._posinputs._replace(
            pos=self.eval((pos[0] + ldisp[0], pos[1] + ldisp[1]))
        )
        self.touchAsset()

    def distanceTo(self, otherpoint):
        """
        Compute the distance to another :class:`_Point` object.

        :param _Point otherpoint: A reference to the other :class:`_Point`
        :rtype: float
        :returns: The distance (in logical units) to the other point
        """
        try:
            return MathApp.distance(self(), otherpoint())
        except AttributeError:
            return otherpoint  # presumably a scalar - use this distance

    @abstractmethod
    def _buildAsset(self):
        pass


class Point(_Point):
    """
    Basic point object representing any point in a geometrical sense.
    An instantiated Point object is *callable* and will return a tuple
    with its logical position as an (x,y) pair.


    :param tuple(float,float) pos: Position in physical or logical units.

    :param \\**kwargs:
        See below

    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' (default) or 'physical'
        * **size** (*int*) Radius of the point (in pixels)
        * **color** (*Color*) Valid :class:`~ggame.asset.Color` object
        * **style** (*LineStyle*) Valid :class:`~ggame.asset.LineStyle` object

    Example:

    .. literalinclude:: ../examples/pointpoint.py

    """

    _defaultsize = 5
    _defaultstyle = LineStyle(0, Color(0, 1))

    def __init__(self, *args, **kwargs):
        super().__init__(
            CircleAsset(self._defaultsize, self._defaultstyle, self._defaultcolor),
            *args,
            **kwargs
        )

    def _buildAsset(self):
        return CircleAsset(
            self._stdinputs.size(), self._stdinputs.style(), self._stdinputs.color()
        )


class ImagePoint(_Point):
    """
    :class:`~ggame.point.Point` object that uses an image as its on-screen
        representation.

    :param str url: Location of an image file (png, jpg)

    :param \\*args:
        See below
    :param \\**kwargs:
        See below

    :Required Arguments:
        * **pos** (*tuple(float,float)*) Position in physical or logical units.


    :Optional Keyword Arguments:
        * **positioning** (*str*) One of 'logical' (default) or 'physical'
        * **frame** (*Frame*) The sub-frame location of image within the image file
        * **qty** (*int*) The number of sub-frames, when used as a sprite sheet
        * **direction** (*str*) One of 'horizontal' (default) or 'vertical'
        * **margin** (*int*) Pixels between sub-frames if sprite sheet
    """

    def __init__(self, url, *args, **kwargs):
        frame = kwargs.get("frame", None)
        qty = kwargs.get("qty", 1)
        direction = kwargs.get("direction", "horizontal")
        margin = kwargs.get("margin", 0)
        self._imageasset = ImageAsset(url, frame, qty, direction, margin)
        super().__init__(self._imageasset, *args, **kwargs)

    def _buildAsset(self):
        return self._imageasset

    def physicalPointTouching(self, ppos):
        """
        Determine if a physical point is considered to be touching point's
        image.

        :param tuple(int,int) ppos: Physical screen coordinates.
        :rtype: boolean
        :returns: True if touching, False otherwise.
        """
        self.setExtents()  # ensure xmin, xmax are correct
        x, y = ppos
        return self.xmax >= x >= self.xmin and self.ymax >= y >= self.ymin
